"""Methods to support Defender play."""
from typing import List, Dict, Tuple, Union
from termcolor import colored

import inspect
from ..logger import log

from bridgeobjects import Card, Suit, CARD_VALUES, SUITS
from bfgsupport import Trick

import bfgcardplay.source.global_variables as global_vars
from .player import Player

MODULE_COLOUR = 'magenta'

def deduce_partner_void_in_trumps(player: Player) -> bool:
    """Return True if we can deduce that partner is void in trumps."""
    suit = player.trump_suit.name
    auction = player.board.auction
    opening_call = auction.seat_calls[player.declarer][0]
    if opening_call.level == 3 and opening_call.denomination == player.trump_suit:
        my_trumps = player.hand.cards_by_suit[suit]
        dummys_trumps = player.hands[player.dummy].cards_by_suit[suit]
        if len(my_trumps) + len(dummys_trumps) >= 5:
            return True
    return False

def dummy_is_short_trump_hand(player: Player) -> bool:
    """Return True if dummy is short trump hand."""
    return len(player.hands[player.dummy].cards_by_suit[player.trump_suit.name]) <= 4

def get_hilo_signal_card(player, cards: List[Card]) -> Card:
    manager = global_vars.manager
    """Return a signal card denoting even/odd"""
    trick = player.board.tricks[-1]
    if len(cards) % 2 == 0:
        if len(cards) == 2:
            if not cards[-2].is_honour:
                manager.set_even_odd(player.seat, trick.suit.name, 0)
                return log(inspect.stack(), cards[-2])
            return log(inspect.stack(), cards[-1])

    manager.set_even_odd(player.seat, trick.suit.name, 1)
    return log(inspect.stack(), cards[-1])

def signal_card(player: Player, manager: object, suit:Suit) -> Union[Card, None]:
    """Signal suit preference."""
    if not manager.signal_card[player.seat]:
        suit_cards = player.suit_cards[suit.name]
        for card in suit_cards:
            if not card.is_honour:
                manager.signal_card[player.seat] = card
                return log(inspect.stack(), card)

        other_suit = player.other_suit_for_signals(suit)
        other_suit_cards = player.suit_cards[other_suit]
        if other_suit_cards:
            manager.signal_card[player.seat] = other_suit_cards[-1]
            return log(inspect.stack(), other_suit_cards[-1])

        for suit_name in SUITS:
            if suit_name != suit.name and suit_name != other_suit:
                final_suit_cards = player.suit_cards[suit_name]
                if final_suit_cards:
                    return log(inspect.stack(), final_suit_cards[-1])
    return None


def surplus_card(player: Player) -> Union[Card, None]:
    # Discard if more cards than the opposition
    for suit_name in SUITS:
        if player.unplayed_cards[suit_name]:
            card_count = len(player.unplayed_cards[suit_name])
            if card_count > len(player.total_unplayed_cards[suit_name]) - card_count:
                return log(inspect.stack(), player.unplayed_cards[suit_name][-1])
    return None

def best_discard(player: Player) -> Card:
    # Discard from longest_suit, but match suit lengths to dummy
    longest_suit = player.longest_suit
    longest_suit_cards = player.suit_cards[longest_suit]
    if len(longest_suit_cards) > len(player.dummys_unplayed_cards[longest_suit]):
        return log(inspect.stack(), longest_suit_cards[-1])

    for suit in SUITS:
        cards = player.suit_cards[suit]
        if len(cards) > len(player.dummys_unplayed_cards[suit]):
            return log(inspect.stack(), cards[-1])

    # Last resort - longest_suit
    return log(inspect.stack(), longest_suit_cards[-1])